/*
 * MRChem, a numerical real-space code for molecular electronic structure
 * calculations within the self-consistent field (SCF) approximations of quantum
 * chemistry (Hartree-Fock and Density Functional Theory).
 * Copyright (C) 2023 Stig Rune Jensen, Luca Frediani, Peter Wind and contributors.
 *
 * This file is part of MRChem.
 *
 * MRChem is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MRChem is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MRChem.  If not, see <https://www.gnu.org/licenses/>.
 *
 * For information on the complete list of contributors to MRChem, see:
 * <https://mrchem.readthedocs.io/>
 */

#pragma once

#include "MRCPP/Plotter"
#include "MRCPP/Printer"

#include "chemistry/Molecule.h"
#include "chemistry/Nucleus.h"

namespace mrchem {

class MolPlotter final : public mrcpp::Plotter<3> {
public:
    MolPlotter(Molecule &mol, const mrcpp::Coord<3> &o)
            : mrcpp::Plotter<3>(o)
            , molecule(&mol) {}

    void linePlot(const std::array<int, 1> &npts, QMFunction &func, const std::string &fname) {
        if (func.hasReal()) linePlot(npts, func.real(), fname + "_re");
        if (func.hasImag()) linePlot(npts, func.imag(), fname + "_im");
    }
    void surfPlot(const std::array<int, 2> &npts, QMFunction &func, const std::string &fname) {
        if (func.hasReal()) surfPlot(npts, func.real(), fname + "_re");
        if (func.hasImag()) surfPlot(npts, func.imag(), fname + "_im");
    }
    void cubePlot(const std::array<int, 3> &npts, QMFunction &func, const std::string &fname) {
        if (func.hasReal()) cubePlot(npts, func.real(), fname + "_re");
        if (func.hasImag()) cubePlot(npts, func.imag(), fname + "_im");
    }

    using Plotter<3>::linePlot;
    using Plotter<3>::surfPlot;
    using Plotter<3>::cubePlot;

protected:
    Molecule *molecule{nullptr};

    void writeCube(const std::array<int, 3> &npts, const Eigen::VectorXd &values) {
        std::ofstream &o = *this->fout;

        auto nNucs = this->molecule->getNNuclei();
        mrcpp::Coord<3> a = calcStep(this->A, npts[0]);
        mrcpp::Coord<3> b = calcStep(this->B, npts[1]);
        mrcpp::Coord<3> c = calcStep(this->C, npts[2]);

        o << "Cube file format" << std::endl;
        o << "Generated by MRChem" << std::endl;

        o.setf(std::ios::scientific);
        o.precision(6);

        // Origin
        o << std::setw(5) << nNucs;
        o << std::setw(15) << this->O[0];
        o << std::setw(15) << this->O[1];
        o << std::setw(15) << this->O[2] << std::endl;

        // Vector A
        o << std::setw(5) << npts[0];
        o << std::setw(15) << a[0];
        o << std::setw(15) << a[1];
        o << std::setw(15) << a[2] << std::endl;

        // Vector B
        o << std::setw(5) << npts[1];
        o << std::setw(15) << b[0];
        o << std::setw(15) << b[1];
        o << std::setw(15) << b[2] << std::endl;

        // Vector C
        o << std::setw(5) << npts[2];
        o << std::setw(15) << c[0];
        o << std::setw(15) << c[1];
        o << std::setw(15) << c[2] << std::endl;

        // Atomic coordinates
        for (auto i = 0; i < nNucs; i++) {
            const auto &nuc = this->molecule->getNuclei()[i];
            const auto &coord = nuc.getCoord();
            auto Z = nuc.getCharge();
            o << std::setw(5) << static_cast<int>(Z);
            o << std::setw(15) << Z;
            o << std::setw(15) << coord[0];
            o << std::setw(15) << coord[1];
            o << std::setw(15) << coord[2] << std::endl;
        }

        // Function values
        o.precision(4);
        for (int n = 0; n < values.size(); n++) {
            o << std::setw(12) << values[n];
            if (n % 6 == 5) o << std::endl; // Line break after 6 values
        }
    }
};

} // namespace mrchem
